package main

import (
	"fmt"
	"os"
	"path/filepath"

	log "github.com/sirupsen/logrus"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"go.salty.im/saltyim"
)

var registerCmd = &cobra.Command{
	Use:     "register nick@domain [broker]",
	Aliases: []string{"reg"},
	Short:   "Creates a new user account and registers it with a Salty Broker",
	Long: `This command registers creates a new identity for a new user, creating a new
key pair and registering with a Salty Broker (in instance of saltyd). Effectively this is
an automated account creation and registration process that combinaes make-user + manual
steps to setup and serve a valid Salty Discovery Endpoint (.well-known/salty/...).

NOTE: This is only spported o a Salty Broker.`,
	Args: cobra.RangeArgs(1, 2),
	Run: func(cmd *cobra.Command, args []string) {
		identity := viper.GetString("identity")
		broker := viper.GetString("broker")

		me, err := saltyim.ParseAddr(args[0])
		if err != nil {
			log.Debugf("error parsing addr: %s\n", err)
		}
		// XXX: What if me.IsZero()

		register(me, identity, broker)
	},
}

func init() {
	rootCmd.AddCommand(registerCmd)

	rootCmd.Flags().StringP(
		"broker", "b", "",
		"use the broker to register with",
	)

	viper.BindPFlag("broker", registerCmd.Flags().Lookup("broker"))
	viper.SetDefault("broker", "")
}

func register(me saltyim.Addr, identity, broker string) {
	dir := filepath.Dir(identity)
	dir = saltyim.FixUnixHome(dir)

	if err := os.MkdirAll(dir, 0700); err != nil {
		fmt.Fprintf(os.Stderr, "error creating configuration directory %s: %s\n", dir, err)
		os.Exit(2)
	}
	identity = filepath.Join(dir, fmt.Sprintf("%s.key", me.User()))

	id, err := saltyim.CreateIdentity(saltyim.WithIdentityPath(identity), saltyim.WithIdentityAddr(me))
	if err != nil {
		fmt.Fprintf(os.Stderr, "error creating identity %q for %s: %s\n", identity, me, err)
		os.Exit(2)
	}

	cli, err := saltyim.NewClient(saltyim.WithIdentity(id))
	if err != nil {
		fmt.Fprintf(os.Stderr, "error initializing client: %s\n", err)
		os.Exit(2)
	}

	if err := cli.Register(broker); err != nil {
		fmt.Fprintf(os.Stderr, "error registering account: %s\n", err)
		os.Exit(2)
	}

	fmt.Println("Success! 🥳")
}
