package main

import (
	"fmt"
	"os"
	"path/filepath"
	"strings"

	"github.com/mitchellh/go-homedir"
	sync "github.com/sasha-s/go-deadlock"
	log "github.com/sirupsen/logrus"
	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"

	"go.salty.im/saltyim"
)

var cfg string

var rootCmd = &cobra.Command{
	Use:     "salty-chat",
	Version: saltyim.FullVersion(),
	Short:   "Salty IM Command-line client",
	Long: `salty.im is an open specification for a new Saltpack based e2e
encrypted messaging protocol and platform for secure communications with
a focus on privacy, security and being self-hosted.

Getting Started:

  $ salty-chat make-user nick@domain
  # follow the instructions

See https://salty.im for more details.`,
	PersistentPreRun: func(cmd *cobra.Command, args []string) {
		// set logging level
		if viper.GetBool("debug") {
			log.SetLevel(log.DebugLevel)
		} else {
			log.SetLevel(log.InfoLevel)

			// Disable deadlock detection in production mode
			sync.Opts.Disable = true
		}
	},
}

// Execute adds all child commands to the root command
// and sets flags appropriately.
// This is called by main.main(). It only needs to happen once to the rootCmd.
func Execute() {
	if len(os.Args) == 2 && (os.Args[1] == "-v" || os.Args[1] == "--version") {
		if err := rootCmd.Execute(); err != nil {
			fmt.Println(err)
			os.Exit(1)
		}
		return
	}

	cmd, _, err := rootCmd.Find(os.Args[1:])
	// default cmd if no cmd is given
	if err == nil && strings.Fields(cmd.Use)[0] == strings.Fields(rootCmd.Use)[0] && cmd.Flags().Parse(os.Args[1:]) != pflag.ErrHelp {
		args := append([]string{strings.Fields(defaultCmd.Use)[0]}, os.Args[1:]...)
		rootCmd.SetArgs(args)
	}

	if err := rootCmd.Execute(); err != nil {
		fmt.Println(err)
		os.Exit(1)
	}
}

func init() {
	cobra.OnInitialize(initConfig)

	rootCmd.PersistentFlags().StringVarP(
		&cfg, "config", "c", "",
		"config file (default is $HOME/.config/salty/config.yml)",
	)

	rootCmd.PersistentFlags().StringP(
		"user", "u", "",
		"User name/profile form config to use",
	)

	rootCmd.PersistentFlags().BoolP(
		"debug", "d", false,
		"Enable debug logging",
	)

	rootCmd.PersistentFlags().StringP(
		"identity", "i", saltyim.DefaultIdentity(),
		"Use the identity file at PATH",
	)

	rootCmd.PersistentFlags().StringP(
		"state", "s", saltyim.DefaultState(),
		"Use the state file at PATH",
	)

	viper.BindPFlag("debug", rootCmd.PersistentFlags().Lookup("debug"))
	viper.SetDefault("debug", false)

	viper.BindPFlag("user", rootCmd.PersistentFlags().Lookup("user"))
	viper.SetDefault("user", "")

	viper.BindPFlag("identity", rootCmd.PersistentFlags().Lookup("identity"))
	viper.SetDefault("identity", saltyim.DefaultIdentity())

	viper.BindPFlag("state", rootCmd.PersistentFlags().Lookup("state"))
	viper.SetDefault("state", saltyim.DefaultState())

	viper.BindPFlag("pre-hook", rootCmd.PersistentFlags().Lookup("pre-hook"))
	viper.SetDefault("pre-hook", "")

	viper.BindPFlag("pre-hook", rootCmd.PersistentFlags().Lookup("pre-hook"))
	viper.SetDefault("post-hook", "")
}

// initConfig reads in config file and ENV variables if set.
func initConfig() {
	if cfg != "" {
		// Use config file from the flag.
		viper.SetConfigFile(cfg)
	} else {
		// Find home directory.
		home, err := homedir.Dir()
		if err != nil {
			fmt.Println(err)
			os.Exit(1)
		}

		viper.AddConfigPath(filepath.Join(home, ".config", "salty"))
		viper.AddConfigPath(filepath.Join("$XDG_CONFIG_HOME", "salty"))

		viper.SetConfigName("config")
		viper.SetConfigType("yml")
	}

	// from the environment
	viper.SetEnvPrefix("SALTY")
	viper.SetEnvKeyReplacer(strings.NewReplacer(".", "_", "-", "_"))
	viper.AutomaticEnv() // read in environment variables that match

	// If a config file is found, read it in.
	if err := viper.ReadInConfig(); err == nil {
		log.Debugf("Using config file: %s", viper.ConfigFileUsed())
	}
}
