package main

import (
	"context"
	"fmt"
	"net/http"
	"net/http/pprof"
	"os"
	"os/signal"
	"path"
	"path/filepath"
	"strings"
	"syscall"

	sync "github.com/sasha-s/go-deadlock"
	log "github.com/sirupsen/logrus"
	flag "github.com/spf13/pflag"
	profiler "github.com/wblakecaldwell/profiler"

	"go.salty.im/saltyim"
	"go.salty.im/saltyim/internal"
)

var (
	bind    string
	debug   bool
	version bool

	// TLS options
	tls     bool
	tlsKey  string
	tlsCert string

	// Basic options
	data          string
	store         string
	baseURL       string
	primaryDomain string

	// Oeprator
	adminUser    string
	supportEmail string
)

const (
	helpText = `
salty.im is an open specification for a new Saltpack based e2e encrypted messaging protocol and platform for secure communications with a focus on privacy, security and being self-hosted.

saltyd is the broker and cross-platform mobile and web app.
`
)

func init() {
	baseProg := filepath.Base(os.Args[0])
	flag.Usage = func() {
		fmt.Fprintf(os.Stderr, "Usage: %s [options]\n", baseProg)
		fmt.Fprint(os.Stderr, helpText)
		flag.PrintDefaults()
	}

	flag.BoolVarP(&debug, "debug", "D", false, "enable debug logging")
	flag.StringVarP(&bind, "bind", "b", "0.0.0.0:8000", "[int]:<port> to bind to")
	flag.BoolVarP(&version, "version", "v", false, "display version information")

	// TLS options
	flag.BoolVar(&tls, "tls", internal.DefaultTLS, "enable TLS (HTTPS)")
	flag.StringVar(&tlsKey, "tls-key", internal.DefaultTLSKey, "path to TLS private key (if blank uses Let's Encrypt)")
	flag.StringVar(&tlsCert, "tls-cert", internal.DefaultTLSCert, "path to TLS certificate (if blank uses Let's Encrypt)")

	// Basic options
	flag.StringVarP(&data, "data", "d", internal.DefaultData, "data directory")
	flag.StringVarP(&store, "store", "s", internal.DefaultStore, "store to use")
	flag.StringVarP(&baseURL, "base-url", "u", internal.DefaultBaseURL, "base url to use")
	flag.StringVarP(&primaryDomain, "primary-domain", "p", internal.DefaultPrimaryDomain, "primary domain delegated to this broker")

	// Oeprator
	flag.StringVarP(&adminUser, "admin-user", "A", internal.DefaultAdminUser, "public key of admin user")
	flag.StringVarP(&supportEmail, "support-email", "E", internal.DefaultSupportEmail, "support email")
}

func flagNameFromEnvironmentName(s string) string {
	s = strings.ToLower(s)
	s = strings.Replace(s, "_", "-", -1)
	return s
}

func parseArgs() error {
	for _, v := range os.Environ() {
		vals := strings.SplitN(v, "=", 2)
		flagName := flagNameFromEnvironmentName(vals[0])
		fn := flag.CommandLine.Lookup(flagName)
		if fn == nil || fn.Changed {
			continue
		}
		if err := fn.Value.Set(vals[1]); err != nil {
			return err
		}
	}
	flag.Parse()
	return nil
}

func main() {
	parseArgs()

	if version {
		fmt.Printf("saltyd %s", saltyim.FullVersion())
		os.Exit(0)
	}

	if debug {
		log.SetLevel(log.DebugLevel)
	} else {
		log.SetLevel(log.InfoLevel)

		// Disable deadlock detection in production mode
		sync.Opts.Disable = true
	}

	svr, err := internal.NewServer(bind,
		// Debug mode
		internal.WithDebug(debug),

		// TLS options
		internal.WithTLS(tls),
		internal.WithTLSKey(tlsKey),
		internal.WithTLSCert(tlsCert),

		// Basic options
		internal.WithData(data),
		internal.WithStore(store),
		internal.WithBaseURL(baseURL),
		internal.WithPrimaryDomain(primaryDomain),

		// Oeprator
		internal.WithAdminUser(adminUser),
		internal.WithSupportEmail(supportEmail),
	)
	if err != nil {
		log.WithError(err).Fatal("error creating server")
	}

	if debug {
		log.Info("starting memory profiler (debug mode) ...")

		go func() {
			// add the profiler handler endpoints
			profiler.AddMemoryProfilingHandlers()

			// start the profiler on service start (optional)
			profiler.StartProfiling()

			// Add pprof handlers
			http.Handle("/debug/pprof/block", pprof.Handler("block"))
			http.Handle("/debug/pprof/goroutine", pprof.Handler("goroutine"))
			http.Handle("/debug/pprof/heap", pprof.Handler("heap"))
			http.Handle("/debug/pprof/threadcreate", pprof.Handler("threadcreate"))

			// listen on port 6060 (pick a port)
			http.ListenAndServe(":6060", nil)
		}()
	}

	ctx, cancel := signal.NotifyContext(context.Background(), syscall.SIGINT, syscall.SIGTERM)
	defer cancel()

	log.Infof("%s v%s listening on http://%s", path.Base(os.Args[0]), saltyim.FullVersion(), bind)
	if err := svr.Run(ctx); err != nil {
		log.WithError(err).Fatal("error running or shutting down server")
	}
}
