// Package authreq signa and verifies HTTP requests using Ed25519 private/public keys
package authreq

import (
	"bytes"
	"context"
	"crypto/ed25519"
	"encoding/base64"
	"fmt"
	"hash/fnv"
	"io"
	"net/http"
	"strings"
	"time"

	"github.com/golang-jwt/jwt/v4"
	"github.com/julienschmidt/httprouter"
	"github.com/oklog/ulid/v2"
	"github.com/patrickmn/go-cache"
	log "github.com/sirupsen/logrus"
)

type contextKey int

const (
	claimsContextKey    contextKey = iota
	authorizationHeader            = "Authorization"
	signatureLifetime              = 5 * time.Minute
)

// ClaimsFromRequest returns the JWT claims object associated with the HTTP request (if any)
func ClaimsFromRequest(r *http.Request) *jwt.RegisteredClaims {
	if claims := r.Context().Value(claimsContextKey); claims != nil {
		return claims.(*jwt.RegisteredClaims)
	}
	return nil
}

// Sign signs the HTTP request with a Ed25519 private key
func Sign(req *http.Request, key ed25519.PrivateKey) (*http.Request, error) {
	pub := enc([]byte(key.Public().(ed25519.PublicKey)))

	h := fnv.New128a()
	fmt.Fprint(h, req.Method, req.URL.Path)

	if req.Body != nil {
		buf := &bytes.Buffer{}
		mw := io.MultiWriter(h, buf)
		_, err := io.Copy(mw, req.Body)
		if err != nil {
			return req, err
		}
		req.Body = io.NopCloser(buf)
	}

	token := jwt.NewWithClaims(jwt.SigningMethodEdDSA, jwt.RegisteredClaims{
		ID:        ulid.Make().String(),
		Subject:   enc(h.Sum(nil)),
		ExpiresAt: jwt.NewNumericDate(time.Now().Add(signatureLifetime)),
		IssuedAt:  jwt.NewNumericDate(time.Now()),
		Issuer:    pub,
	})

	sig, err := token.SignedString(key)
	if err != nil {
		return req, err
	}

	req.Header.Set(authorizationHeader, sig)

	return req, nil
}

// VerifyMiddleware is a httprouter.handler middleware that verifies HTTP requests previously
// signed with a Ed25519 private key
func VerifyMiddleware(next httprouter.Handle) httprouter.Handle {
	usedIDs := cache.New(signatureLifetime, 2*signatureLifetime)

	return func(rw http.ResponseWriter, req *http.Request, p httprouter.Params) {
		auth := req.Header.Get(authorizationHeader)
		if auth == "" {
			log.Warn("no authorization found")
			rw.WriteHeader(http.StatusUnauthorized)
			return
		}

		hash := fnv.New128a()
		fmt.Fprint(hash, req.Method, req.URL.Path)

		if req.Body != nil {
			buf := &bytes.Buffer{}
			mw := io.MultiWriter(hash, buf)
			_, err := io.Copy(mw, req.Body)
			if err != nil {
				log.WithError(err).Error("error hashing request body")
				rw.WriteHeader(http.StatusBadRequest)
				return
			}
			req.Body = io.NopCloser(buf)
		}

		subject := enc(hash.Sum(nil))
		token, err := jwt.ParseWithClaims(
			string(auth),
			&jwt.RegisteredClaims{},
			func(token *jwt.Token) (any, error) {
				claims, ok := token.Claims.(*jwt.RegisteredClaims)
				if !ok {
					return nil, fmt.Errorf("wrong type of claim")
				}

				pub, err := dec(claims.Issuer)
				return ed25519.PublicKey(pub), err
			},
			jwt.WithValidMethods([]string{"EdDSA"}),
			jwt.WithJSONNumber(),
		)
		if err != nil {
			log.WithError(err).Error("error verifying token")
			rw.WriteHeader(http.StatusBadRequest)
			return
		}

		claims, ok := token.Claims.(*jwt.RegisteredClaims)
		if !ok {
			log.Warn("no claims found")
			rw.WriteHeader(http.StatusUnprocessableEntity)
			return
		}

		if claims.ID != "" {
			if _, ok := usedIDs.Get(claims.ID); ok {
				log.Warnf("ID has been seen before: %s", claims.ID)
				rw.WriteHeader(http.StatusForbidden)
				return
			}
			usedIDs.SetDefault(claims.ID, true)
		}

		if claims.Subject != subject {
			log.Warnf("subjects do not match %s != %q", claims.Subject, subject)
			rw.WriteHeader(http.StatusForbidden)
			return
		}

		ctx := context.WithValue(req.Context(), claimsContextKey, claims)
		next(rw, req.WithContext(ctx), p)
	}
}

func enc(b []byte) string {
	return base64.RawURLEncoding.EncodeToString(b)
}
func dec(s string) ([]byte, error) {
	s = strings.TrimSpace(s)
	return base64.RawURLEncoding.DecodeString(s)
}
